@@ -1,17 +1,19 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from collections import defaultdict
 import datetime
+from collections import defaultdict
+from itertools import groupby
 
-from sql import Null
+from sql import Literal, Null
 
-from trytond.model import Model, ModelView, ModelSQL, fields, Unique, \
-    sequence_ordered
-from trytond.pyson import If, Eval
-from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
-    Button
+from trytond.model import (
+    Index, Model, ModelSQL, ModelView, Unique, fields, sequence_ordered)
+from trytond.modules.currency.fields import Monetary
 from trytond.pool import Pool
+from trytond.pyson import Eval
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
 
 class Procedure(ModelSQL, ModelView):
@@ -30,14 +32,16 @@ class Procedure(ModelSQL, ModelView):
 class Level(sequence_ordered(), ModelSQL, ModelView):
     'Account Dunning Level'
     __name__ = 'account.dunning.level'
-    procedure = fields.Many2One('account.dunning.procedure', 'Procedure',
-        required=True, select=True)
+    procedure = fields.Many2One(
+        'account.dunning.procedure', "Procedure", required=True)
     overdue = fields.TimeDelta('Overdue',
         help="When the level should be applied.")
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
         table = cls.__table_handler__(module_name)
         sql_table = cls.__table__()
 
@@ -47,9 +51,9 @@ class Level(sequence_ordered(), ModelSQL, ModelView):
         if table.column_exist('days'):
             cursor.execute(*sql_table.select(sql_table.id, sql_table.days,
                     where=sql_table.days != Null))
-            for id_, days in cursor.fetchall():
+            for id_, days in cursor:
                 overdue = datetime.timedelta(days)
-                cursor.execute(*sql_table.update(
+                update.execute(*sql_table.update(
                         [sql_table.overdue],
                         [overdue],
                         where=sql_table.id == id_))
@@ -67,19 +71,14 @@ class Level(sequence_ordered(), ModelSQL, ModelView):
 _STATES = {
     'readonly': Eval('state') != 'draft',
     }
-_DEPENDS = ['state']
 
 
 class Dunning(ModelSQL, ModelView):
     'Account Dunning'
     __name__ = 'account.dunning'
-    company = fields.Many2One('company.company', 'Company', required=True,
-        help="Make the dunning belong to the company.",
-        select=True, domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
-        states=_STATES, depends=_DEPENDS)
+    company = fields.Many2One(
+        'company.company', "Company", required=True, states=_STATES,
+        help="Make the dunning belong to the company.")
     line = fields.Many2One('account.move.line', 'Line', required=True,
         help="The receivable line to dun for.",
         domain=[
@@ -90,41 +89,59 @@ class Dunning(ModelSQL, ModelView):
                 ('credit', '<', 0),
                 ],
             ],
-        states=_STATES, depends=_DEPENDS + ['company'])
+        states=_STATES)
     procedure = fields.Many2One('account.dunning.procedure', 'Procedure',
-        required=True, states=_STATES, depends=_DEPENDS)
+        required=True, states=_STATES)
     level = fields.Many2One('account.dunning.level', 'Level', required=True,
         domain=[
             ('procedure', '=', Eval('procedure', -1)),
             ],
-        states=_STATES, depends=_DEPENDS + ['procedure'])
+        states=_STATES)
+    date = fields.Date(
+        "Date", readonly=True,
+        states={
+            'invisible': Eval('state') == 'draft',
+            },
+        help="When the dunning reached the level.")
+    age = fields.Function(fields.TimeDelta(
+            "Age",
+            states={
+                'invisible': Eval('state') == 'draft',
+            },
+            help="How long the dunning has been at the level."),
+        'get_age')
     blocked = fields.Boolean('Blocked',
         help="Check to block further levels of the procedure.")
     state = fields.Selection([
             ('draft', 'Draft'),
             ('waiting', "Waiting"),
             ('final', "Final"),
-            ], 'State', readonly=True)
+            ], 'State', readonly=True, sort=False)
     active = fields.Function(fields.Boolean('Active'), 'get_active',
         searcher='search_active')
-    party = fields.Function(fields.Many2One('party.party', 'Party'),
+    party = fields.Function(fields.Many2One(
+            'party.party', 'Party',
+            context={
+                'company': Eval('company', -1),
+                },
+            depends={'company'}),
         'get_line_field', searcher='search_line_field')
-    amount = fields.Function(fields.Numeric('Amount',
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    amount = fields.Function(Monetary(
+            "Amount", currency='currency', digits='currency'),
         'get_amount')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'get_line_field')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"), 'get_line_field')
     maturity_date = fields.Function(fields.Date('Maturity Date'),
         'get_line_field', searcher='search_line_field')
-    amount_second_currency = fields.Function(fields.Numeric(
+    amount_second_currency = fields.Function(Monetary(
             'Amount Second Currency',
-            digits=(16, Eval('second_currency_digits', 2)),
-            depends=['second_currency_digits']), 'get_amount_second_currency')
+            currency='second_currency', digits='second_currency',
+            states={
+                'invisible': Eval('currency') == Eval('second_currency'),
+                }),
+        'get_amount_second_currency')
     second_currency = fields.Function(fields.Many2One('currency.currency',
             'Second Currency'), 'get_second_currency')
-    second_currency_digits = fields.Function(fields.Integer(
-            'Second Currency Digits'), 'get_second_currency_digits')
 
     @classmethod
     def __setup__(cls):
@@ -134,7 +151,18 @@ class Dunning(ModelSQL, ModelView):
             ('line_unique', Unique(table, table.line),
                 'account_dunning.msg_dunning_line_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                table,
+                (table.state, Index.Equality()),
+                where=table.state.in_(['draft', 'waiting'])))
         cls._active_field = 'active'
+        cls._buttons.update({
+                'reschedule': {
+                    'invisible': ~Eval('active', True),
+                    'depends': ['active'],
+                    },
+                })
 
     @classmethod
     def __register__(cls, module):
@@ -155,6 +183,20 @@ class Dunning(ModelSQL, ModelView):
     def default_state():
         return 'draft'
 
+    def get_age(self, name):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        if self.date:
+            with Transaction().set_context(company=self.company.id):
+                return Date.today() - self.date
+
+    @classmethod
+    def order_age(cls, tables):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        table, _ = tables[None]
+        return [Literal(Date.today()) - table.date]
+
     def get_active(self, name):
         return not self.line.reconciliation
 
@@ -185,12 +227,6 @@ class Dunning(ModelSQL, ModelView):
         else:
             return self.line.account.company.currency.id
 
-    def get_second_currency_digits(self, name):
-        if self.line.second_currency:
-            return self.line.second_currency.digits
-        else:
-            return self.line.account.company.currency.digits
-
     @classmethod
     def search_active(cls, name, clause):
         reverse = {
@@ -229,10 +265,13 @@ class Dunning(ModelSQL, ModelView):
             date = Date.today()
 
         set_level = defaultdict(list)
-        for dunning in cls.search([
+        with Transaction().set_context(_check_access=True):
+            dunnings = cls.search([
                     ('state', '=', 'waiting'),
                     ('blocked', '=', False),
-                    ]):
+                    ])
+        dunnings = cls.browse(dunnings)
+        for dunning in dunnings:
             procedure = dunning.procedure
             levels = procedure.levels
             levels = levels[levels.index(dunning.level) + 1:]
@@ -252,15 +291,19 @@ class Dunning(ModelSQL, ModelView):
                 to_write.extend((dunnings, {
                             'level': level.id,
                             'state': 'draft',
+                            'date': None,
                             }))
             else:
                 to_write.extend((dunnings, {
                             'state': 'final',
+                            'date': Date.today(),
                             }))
         if to_write:
             cls.write(*to_write)
 
-        lines = MoveLine.search(cls._overdue_line_domain(date))
+        with Transaction().set_context(_check_access=True):
+            lines = MoveLine.search(cls._overdue_line_domain(date))
+        lines = MoveLine.browse(lines)
         dunnings = (cls._get_dunning(line, date) for line in lines)
         cls.save([d for d in dunnings if d])
 
@@ -282,10 +325,21 @@ class Dunning(ModelSQL, ModelView):
 
     @classmethod
     def process(cls, dunnings):
-        cls.write([d for d in dunnings
-                if not d.blocked and d.state == 'draft'], {
-                'state': 'waiting',
-                })
+        pool = Pool()
+        Date = pool.get('ir.date')
+        for company, c_dunnings in groupby(dunnings, key=lambda d: d.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            cls.write([d for d in c_dunnings
+                    if not d.blocked and d.state == 'draft'], {
+                    'state': 'waiting',
+                    'date': today,
+                    })
+
+    @classmethod
+    @ModelView.button_action('account_dunning.act_reschedule_dunning_wizard')
+    def reschedule(cls, dunnings):
+        pass
 
 
 class CreateDunningStart(ModelView):
@@ -351,3 +405,15 @@ class ProcessDunning(Wizard):
     def transition_process(self):
         self.model.process(self.records)
         return self.next_state('process')
+
+
+class RescheduleDunning(Wizard):
+    "Reschedule Account Dunning"
+    __name__ = 'account.dunning.reschedule'
+    start = StateAction('account.act_reschedule_lines_wizard')
+
+    def do_start(self, action):
+        return action, {
+            'ids': [self.record.line.id],
+            'model': 'account.move.line',
+            }
